/********************************************************************************************************
* PROGRAM      : FreeGo4
* AUTHOR       : Ferry Guillaume (guillaume@freego.fr), AZEVEDO Filipe (filipe@freego.fr)
* FILENAME     : UITelevision.cpp
* LICENSE      : GPL
* COMMENTARY   : 
********************************************************************************************************/
#include <QFileDialog>
#include <QProcess>

#include "main.h"
#include "UITelevision.h"
#include "ui/UIMain.h"
#include "misc.h"
#include "Options.h"

#include <QDebug>

QRegExp rxAvi( "(\\d\\d\\d)_(.+)_(\\d\\d)(\\d\\d)(\\d\\d\\d\\d)_(\\d\\dh\\d\\d)(_(.+))?.avi" );

QString getLogo( const QString& url )
{ return QString( ":/television/Pictures/television/logos/%1.png" ).arg( url.section( "/", -1, -1 ) ); }

QString getChannelId( const QString& url )
{ return url.section( "/", -1, -1 ); }

QString getLogoFromChannelId( const QString& id )
{ return getLogo( QString( "/%1" ).arg( id ) ); }

// --- STATIC POINTER ---
QPointer<UITelevision> UITelevision::mSelf = 0L;
UITelevision* UITelevision::self( QWidget* parent )
{
	if ( !mSelf )
		mSelf = new UITelevision( parent );
	return mSelf;
}

// --- UITELEVISION ---
UITelevision::UITelevision( QWidget* inWidget )
	: fgDialog( inWidget )
{
	mMonths << "Janvier" << QString::fromUtf8( "Février" ) << "Mars" << "Avril" << "Mai" << "Juin" << "Juillet" << QString::fromUtf8( "Août" ) << "Septembre" << "Octobre" << "Novembre" << QString::fromUtf8( "Décembre" );
	setupUi( this );
	fixMacWidgets();
	initObjects();
	initSignals();
	enableDays( false );
	fgMenuList->setToggled( true );
	restoreFavoritesChannels();
	syncRecordsFolder();
}

UITelevision::~UITelevision()
{ saveFavoritesChannels(); }

// --- INITOBJECTS ---
void UITelevision::initObjects()
{
	// Inits
	grabber = new ChannelsGrabber();
	
	// VLC Check
	checkVLC();

	// Settings
	recordPath = Options::current()->value("/FreeGo/Options/RecordPath").toString();
	fgDirLabel->setText( recordPath );
	
	cbFullscreen->setToggled( Options::current()->value("/FreeGo/Options/Fullscreen").toBool() );
	cbOnTop->setToggled( Options::current()->value("/FreeGo/Options/OnTop").toBool() );
	cbShowVideo->setToggled( Options::current()->value("/FreeGo/Options/ShowVideo").toBool() );
	
	// Channels Tree : Columns Size
	fgList->setColumnWidth(0, 40);
	fgList->setColumnWidth(1, 310);
	fgList->setColumnWidth(2, 40);
	
	// Records Tree : Columns Size
	fgTree->setColumnWidth( 0, 280 ); // logo/chaine
	fgTree->setColumnWidth( 1, 100 ); // etat
	
	// Convenience Lists
	days << fgMonday << fgTuesday << fgWednesday << fgThursday << fgFriday << fgSaturday << fgSunday;
	
	// Default Values
	fgDateEdit->setDate( QDate::currentDate() );
	fgStartEdit->setTime( QTime::currentTime() );
	fgEndEdit->setTime( QTime::currentTime() );
	
	// Getting Records :)
	for(int i=0; i< Recorder::self()->getNbRecords(); i++)
		addRecord( Recorder::self()->getRecord(i) );
	
	Recorder::self()->checkRecords();
}

// --- INITSIGNALS ---
void UITelevision::initSignals()
{
	// menu
	QList<fgLabel*> buttons = findChildren<fgLabel*>( QRegExp( "fgMenu" ) );
	foreach ( fgLabel* button, buttons )
		connect( button, SIGNAL( clicked( const int ) ), this, SLOT( fgMenu_clicked( const int ) ) );
	connect(grabber, 			SIGNAL(completed()), 				this, SLOT(setChannels())); 
	connect(Recorder::self(), 	SIGNAL(newRecord(Record*)), 		this, SLOT(addRecord(Record*)));
	connect(Recorder::self(), 	SIGNAL(changeState(int, int)), 		this, SLOT(updateState(int, int)));
	connect(Recorder::self(),	SIGNAL(recordFinished(Record*)),	this, SLOT(addToFinished(Record*)));
	connect(fgWeekly,			SIGNAL(toggled(bool)), 				this, SLOT(enableDays(bool)));
	connect(fgWeek,				SIGNAL(toggled(bool)),				this, SLOT(toggleDays(bool)));
	connect(fgBrowseVLC,		SIGNAL(clicked()),					this, SLOT(browseForVLC()));
	connect(fgValidVLC,			SIGNAL(clicked()),					this, SLOT(checkVLCPath()));
	connect( fgFavTree, SIGNAL( itemDoubleClicked( QListWidgetItem* ) ), this, SLOT( on_fgFavWatch_clicked() ) );
	
	QAction* action = new QAction( QString::fromUtf8( "Changer l'état favoris des items" ), this );
	fgList->addAction( action );
	connect( action, SIGNAL( triggered() ), this, SLOT( addRemoveFavorites() ) );
}

// --- ENABLEDAYS ---
void UITelevision::enableDays(bool flag)
{
	fgWeek->setEnabled(flag);
	for(int i=0; i<days.count(); i++)
		days.at(i)->setEnabled(flag);
}

// --- TOGGLEDAYS ---
void UITelevision::toggleDays(bool flag)
{
	for(int i=0; i<days.count(); i++)
		days.at(i)->setToggled(flag);
}

// --- ADDRECORD ---
void UITelevision::addRecord(Record *inRec)
{
	QTreeWidgetItem* it = new QTreeWidgetItem( fgTree );
	it->setData( 0, Qt::UserRole, reinterpret_cast<quintptr>( inRec ) );
	updateState( fgTree->indexOfTopLevelItem( it ), 0 );
}

// --- UPDATESTATE ---
void UITelevision::updateState(int idx, int state)
{
	if ( QTreeWidgetItem* it = fgTree->topLevelItem( idx ) )
	{
		Record* rec = reinterpret_cast<Record*>( it->data( 0, Qt::UserRole ).value<quintptr>() );
		switch ( state )
		{
			case 0:
			{
				QString tooltip;
				QString programmation = "Unique";
				if ( rec->weekly )
				{
					if ( rec->days != "*" )
					{
						QStringList mDays;
						foreach ( const QChar& c, rec->days )
							mDays << days.at( c.digitValue() -1 )->text();
						programmation = QString( "Hebdomaraire (%1)" ).arg( mDays.join( ", " ) );
					}
					else
						programmation = "Hebdomaraire (Tout les jours)";
				}
				else
					programmation += QString( " le %1" ).arg( rec->date.toString( "dd/MM/yyyy" ) );
				programmation += QString::fromUtf8( " de %1 à %2" ).arg( rec->hstart.toString( "hh:mm" ) ).arg( rec->hstop.toString( "hh:mm" ) );
				it->setIcon( 0, QIcon( getLogo( rec->url ) ) );
				it->setIcon( 1, QIcon( ":/television/pause.png" ) );
				it->setText( 0, rec->channel );
				it->setText( 1, "En attente" );
				it->setToolTip( 0, programmation );
				it->setToolTip( 1, programmation );
				break;
			}
			case 1:
			{
				it->setIcon( 1, QIcon( ":/television/rec.png" ) );
				it->setText( 1, "En cours" );
				break;
			}
			case 2:
			{
				it->setIcon( 1, QIcon( ":/television/stop.png" ) );
				it->setText( 1, QString::fromUtf8( "Terminé" ) );
				break;
			}
		}
	}
}

// --- ADDTOFINISHED ---
void UITelevision::addToFinished(Record *record)
{
	const int id = Recorder::self()->getId( record );
	Recorder::self()->removeRecord( id );
	delete fgTree->topLevelItem( id );
	syncRecordsFolder();
}

// --- Stack Widget Controls ---
// --- FGBUTTONS_CLICKED ---
void UITelevision::fgMenu_clicked( const int id )
{
	QList<fgLabel*> buttons = findChildren<fgLabel*>( QRegExp( "fgMenu" ) );
	foreach ( fgLabel* button, buttons )
	{
		button->setToggled( button->data().toInt() == id );
		if ( sender() && button->data().toInt() == id )
			button->setToggled( false );
	}
	fgTVStack->setCurrentIndex( id );
}

// --- SETCHANNELS ---
void UITelevision::setChannels()
{
	QList<Channel> channels = grabber->getChannels();
	QTreeWidgetItem* item;
	
	for ( int i = 0; i < channels.count(); i++ )
	{
		item = new QTreeWidgetItem( fgList );
		item->setText( 1, channels.at( i ).name );
		item->setData( 1, Qt::UserRole +1, channels.at( i ).url );
		if ( fgFavTree->findItems( channels.at( i ).name, Qt::MatchFixedString | Qt::MatchWrap ).count() > 0 )
			item->setIcon( 2, QPixmap( ":/television/Pictures/television/star.png" ) );
		item->setIcon( 0, QIcon( getLogo( channels.at(i).url ) ) );
		// ComboBox
		fgChooseBox->addItem( channels.at(i).name, channels.at(i).url );
	}
}

// --- CHECKVLC ---
void UITelevision::checkVLC()
{	
	// Television
	vlcPath = Options::current()->value("/FreeGo/Options/VLCpath").toString();
	
	if ( vlcPath.size() <= 0 )
	{
#ifdef Q_OS_WIN
		QSettings settings( "\\HKEY_LOCAL_MACHINE\\SOFTWARE\\VideoLAN\\VLC", QSettings::NativeFormat );
		vlcPath = settings.value( "InstallDir" ).toString();
		if (vlcPath.size() > 0)
		{
			vlcPath.append("\\vlc.exe");
			vlcPath.prepend( "\"" ).append( "\"" );
		}
#elif defined( Q_OS_MAC )
		vlcPath = "/Applications/VLC.app";
#else
		vlcPath = "vlc";
#endif
		if ( !vlcPath.isEmpty() )
			Options::current()->setValue("/FreeGo/Options/VLCpath", vlcPath);
	}
	fgVLCPath->setText(vlcPath);
}

// --- CHECKVLCPATH ---
void UITelevision::checkVLCPath()
{
#ifdef Q_OS_WIN
	if ( fgVLCPath->text().endsWith( "vlc.exe" ) )
#elif defined Q_OS_MAC
	if ( fgVLCPath->text().endsWith( "VLC.app" ) )
#else
	if ( fgVLCPath->text().endsWith( "vlc" ) )
#endif
	{
		vlcPath = fgVLCPath->text();
		UIMAIN->alert( QString( "L'emplacement %1 pour VLC est correct!" ).arg( vlcPath ) );
	}
	else
	{
		UIMAIN->alert( QString( "L'emplacement %1 pour VLC est incorrect!" ).arg( vlcPath ) );
	}
}

// --- BROWSEFORVLC ---
void UITelevision::browseForVLC()
{
	QString fileName = QFileDialog::getOpenFileName(this,
							QString::fromUtf8( "Sélectionnez l'emplacement de VLC" ),
							QDir::homePath(),
#ifdef Q_OS_WIN
							QString::fromUtf8( "Exécutable VLC (vlc.exe);;Tous les fichiers (*)" )
#elif defined Q_OS_MAC
							QString::fromUtf8( "Bundle VLC (VLC.app);;Tous les fichiers (*)" )
#else
							QString::fromUtf8( "Exécutable VLC (vlc);;Tous les fichiers (*)" )
#endif
							);
	if (!fileName.isEmpty())
	{
		fgVLCPath->setText(fileName);
		checkVLCPath();
	}
}

bool UITelevision::watchUrl( const QUrl& url )
{
	if ( url.isEmpty() )
		return false;
	const bool fullscreen = Options::current()->value( "/FreeGo/Options/Fullscreen", false ).toBool();
	const bool onTop = Options::current()->value( "/FreeGo/Options/OnTop", false ).toBool();
	const QString command = QString( "\"%1\" %2 %3 \"%4\"" )
		.arg( vlcPath
#ifdef Q_OS_MAC
			+"/Contents/MacOS/VLC"
#endif
		)
		.arg( fullscreen ? "--fullscreen" : "" )
		.arg( onTop ? "--video-on-top" : "" )
		.arg( url.toString() );
	qWarning() << command;
	return QProcess::startDetached( command );
}

bool UITelevision::recordNow( const QUrl& url, const QString& channel )
{
	if ( url.isEmpty() )
		return false;
	const bool showVideo = Options::current()->value( "/FreeGo/Options/ShowVideo", false ).toBool();
	const bool fullscreen = Options::current()->value( "/FreeGo/Options/Fullscreen", false ).toBool();
	const bool onTop = Options::current()->value( "/FreeGo/Options/OnTop", false ).toBool();
	const QString fn = QDir::cleanPath( QString( "%1/%2_%3_%4_%5.avi" )
		.arg( recordPath )
		.arg( getChannelId( url.toString() ) )
		.arg( channel.simplified().replace( " ", "_" ) )
		.arg( QDate::currentDate().toString( "ddMMyyyy" ) )
		.arg( QTime::currentTime().toString( "hh'h'mm" ) ).replace( "\\", "/" ) );
	QString command;
	if ( showVideo )
		command = QString( "\"%1\" %2 %3--verbose 0 --sout \"#duplicate{dst=display,dst=std{mux=ts,access=file,dst=%4}}\" \"%5\"" )
		.arg( vlcPath
#ifdef Q_OS_MAC
			+"/Contents/MacOS/VLC"
#endif
		)
		.arg( fullscreen ? "--fullscreen" : "" )
		.arg( onTop ? "--video-on-top" : "" )
		.arg( fn )
		.arg( url.toString() );
	else
		command = QString( "\"%1\" --verbose 0 --sout file/ts:\"%2\" \"%3\"" )
		.arg( vlcPath
#ifdef Q_OS_MAC
			+"/Contents/MacOS/VLC"
#endif
		)
		.arg( fn )
		.arg( url.toString() );
	return QProcess::startDetached( command );
}

// --- ON_FGLIST_ITEMDOUBLECLICKED ---
void UITelevision::on_fgList_itemDoubleClicked(QTreeWidgetItem *item, int column)
{
	if ( !item )
		return;
	if ( column < 2 )
		watchUrl( QUrl( item->data( 1, Qt::UserRole +1 ).toString() ) );
	else
		addRemoveFavorites();
}

// --- ON_FGWATCH_CLICKED ---
void UITelevision::on_fgWatch_clicked()
{
	QTreeWidgetItem* item = fgList->selectedItems().value( 0 );
	if ( item )
		watchUrl( QUrl( item->data( 1, Qt::UserRole +1 ).toString() ) );
}

// --- ON_FGRECORDNOW_CLICKED ---
void UITelevision::on_fgRecordNow_clicked()
{
	QTreeWidgetItem* item = fgList->selectedItems().value( 0 );
	if ( item )
		recordNow( item->data( 1, Qt::UserRole + 1 ).toString(), item->text( 1 ) );
}

// --- ON_FGFAVWATCH_CLICKED ---
void UITelevision::on_fgFavWatch_clicked()
{
	QListWidgetItem* item = fgFavTree->selectedItems().value( 0 );
	if ( item )
		watchUrl( QUrl( item->data( Qt::UserRole +1 ).toString() ) );
}

// --- ON_FGREMOVECHANNEL_CLICKED ---
void UITelevision::on_fgRemoveChannel_clicked()
{
	foreach ( QListWidgetItem* item, fgFavTree->selectedItems() )
	{
		QList<QTreeWidgetItem*> its = fgList->findItems( item->text(), Qt::MatchFixedString | Qt::MatchWrap | Qt::MatchRecursive, 1 );
		foreach ( QTreeWidgetItem* it, its )
			it->setIcon( 2, QIcon() );
		delete item;
	}
}	

// --- ON_FGREMOVESEL_CLICKED ---
void UITelevision::on_fgRemoveSel_clicked()
{
	QList<QTreeWidgetItem*> items = fgTree->selectedItems();
	
	for(int i=0; i<items.count(); i++)
	{
		const int id = fgTree->indexOfTopLevelItem( items.at( i ) );
		Recorder::self()->removeRecord( id );
		delete fgTree->topLevelItem( id );
	}
}

// --- ON_FGADDFAVORITE_CLICKED ---
void UITelevision::addRemoveFavorites()
{
	fgFavTree->clearSelection();
	// get items
	QList<QTreeWidgetItem*> items = fgList->selectedItems();
	// check items
	foreach ( QTreeWidgetItem* item, items )
	{
		if ( !item->icon( 2 ).isNull() )
		{
			items.removeAll( item );
			QList<QListWidgetItem*> selitems = fgFavTree->findItems( item->text( 1 ), Qt::MatchFixedString | Qt::MatchWrap );
			foreach ( QListWidgetItem* it, selitems )
				it->setSelected( true );
		}
	}
	// remove items
	on_fgRemoveChannel_clicked();
	// add items
	foreach ( QTreeWidgetItem* it, items )
	{
		QListWidgetItem* item = new QListWidgetItem( fgFavTree );
		item->setText( it->text( 1 ) );
		item->setIcon( it->icon( 0 ) );
		item->setData( Qt::UserRole +1, it->data( 1, Qt::UserRole +1 ).toString() );
		it->setIcon( 2, QPixmap( ":/television/Pictures/television/star.png" ) );
	}
}

// --- ON_FGEDIT_CLICKED ---
void UITelevision::on_fgEdit_clicked()
{}

// --- ON_FGSETPATH_CLICKED ---
void UITelevision::on_fgSetPath_clicked()
{
	QString filePath = QFileDialog::getSaveFileName(this,
										"Veuillez saisir un nom de fichier",
										recordPath,
										QString::fromUtf8("Clip Vidéo AVI (*.avi)"));
	if (!filePath.isEmpty())
		fgPath->setText( filePath );
}

// --- ON_FGCONFIRM_CLICKED ---
void UITelevision::on_fgConfirm_clicked()
{	
	Record *record 	= new Record;

	record->hstart 	= fgStartEdit->time();
	record->hstop	= fgEndEdit->time();
	record->channel	= fgChooseBox->currentText();
	record->url	= fgChooseBox->itemData(fgChooseBox->currentIndex()).toString();
	
	// Date / Days Settings
	if (fgSingleShot->isChecked())
	{
		record->weekly 	= 0;
		record->date 	= fgDateEdit->date();
	}
	else
	{
		QString strDays;
		QString num;
		
		// Date
		if (fgWeek->toggled())
			strDays = "*";		
		else
			for(int i=0; i<days.count(); i++)
				if ( days.at(i)->toggled() )
					strDays.append( num.setNum(i+1) );
		
		record->weekly 	= 1;
		record->days 	= strDays;
	}
	
	// Path
	if ( fgAutoDir->toggled() )
		record->path = QString( "%1/%2_%3_%4_%5.avi" )
					.arg( recordPath )
					.arg( getChannelId( record->url ) )
					.arg( record->channel.simplified().replace(" ", "_") )
					.arg( record->date.toString("ddMMyyyy") )
					.arg( record->hstart.toString("hh'h'mm") );
	else
		record->path = fgPath->text();	// Check If Path / Filename are valid
	
	Recorder::self()->addRecord( record );
}

// --- ON_FGBROWSE_CLICKED ---
void UITelevision::on_fgBrowse_clicked()
{
	QString path = QFileDialog::getExistingDirectory(this,
									QString::fromUtf8("Sélectionnez un répertoire"),
								recordPath,
								QFileDialog::ShowDirsOnly);
	if ( !path.isEmpty() )
	{
		recordPath = path;
		fgDirLabel->setText( recordPath );
		Options::current()->setValue("/FreeGo/Options/RecordPath", recordPath);
	}
}

// --- ON_FGFINISHEDTREE_ITEMDOUBLECLICKED ---
void UITelevision::on_fgFinishedTree_itemDoubleClicked(QTreeWidgetItem*, int)
{ on_fgFinishedOpen_clicked(); }

void UITelevision::on_fgFinishedRemove_clicked()
{
	QTreeWidgetItem* item = fgFinishedTree->selectedItems().value( 0 );
	if ( item )
	{
		if ( QFile::remove( item->data( 0, Qt::UserRole ).toString() ) )
			delete item;
	}
}

void UITelevision::on_fgFinishedOpen_clicked()
{
	QTreeWidgetItem* item = fgFinishedTree->selectedItems().value( 0 );
	if ( item )
		watchUrl( QUrl( item->data( 0, Qt::UserRole ).toString() ) );
}

void UITelevision::restoreFavoritesChannels()
{
	QSettings* settings = Options::current();
	const int count = settings->beginReadArray( "/FreeGo/FavoritesChannels" );
	for ( int i = 0; i < count; i++ )
	{
		settings->setArrayIndex( i );
		QListWidgetItem* item = new QListWidgetItem( fgFavTree );
		item->setIcon( QIcon( getLogo( settings->value( "Url" ).toString() ) ) );
		item->setText( settings->value( "Text" ).toString() );
		item->setData( Qt::UserRole +1, settings->value( "Url" ).toString() );
	}
	settings->endArray();
}

void UITelevision::saveFavoritesChannels()
{
	QSettings* settings = Options::current();
	const int count = fgFavTree->count();
	settings->beginWriteArray( "/FreeGo/FavoritesChannels", count );
	for ( int i = 0; i < count; i++ )
	{
		QListWidgetItem* item = fgFavTree->item( i );
		settings->setArrayIndex( i );
		settings->setValue( "Text", item->text() );
		settings->setValue( "Url", item->data( Qt::UserRole +1 ).toString() );
	}
	settings->endArray();
}

void UITelevision::syncRecordsFolder()
{
	fgFinishedTree->clear();
	QDir dir( recordPath );
	QFileInfoList files = dir.entryInfoList( QStringList( "*.avi" ), QDir::Files, QDir::Name );
	foreach ( const QFileInfo& fi, files )
	{
		QTreeWidgetItem* it = new QTreeWidgetItem( fgFinishedTree );
		if ( rxAvi.exactMatch( fi.fileName() ) )
		{
			const QString channelId = rxAvi.cap( 1 );
			const QString channel = rxAvi.cap( 2 ).replace( "_", " " );
			const QString month = mMonths.value( rxAvi.cap( 4 ).toInt() -1 );
			const QString name = rxAvi.cap( 7 ).replace( "_", " " ).replace( ".", " " ).simplified();
			QString text = QString::fromUtf8( "%1 le %2 %3 %4 à %5" ).arg( channel ).arg( rxAvi.cap( 3 ) ).arg( month ).arg( rxAvi.cap( 5 ) ).arg( rxAvi.cap( 6 ) );
			if ( !name.isEmpty() )
				text.append( QString( " (%1)" ).arg( name ) );
			it->setIcon( 0, QIcon( getLogoFromChannelId( channelId ) ) );
			it->setText( 0, text );
		}
		else
		{
			it->setIcon( 0, QIcon( getLogoFromChannelId( "nologo" ) ) );
			it->setText( 0, fi.completeBaseName().replace( "_", " " ).replace( ".", " " ).simplified() );
		}
		it->setData( 0, Qt::UserRole, fi.absoluteFilePath() );
	}
}

void UITelevision::on_fgApply_clicked()
{
	Options::current()->setValue( "/FreeGo/Options/VLCpath", fgVLCPath->text() );
	Options::current()->setValue( "/FreeGo/Options/Fullscreen", cbFullscreen->toggled() );
	Options::current()->setValue( "/FreeGo/Options/OnTop", cbOnTop->toggled() );
	Options::current()->setValue( "/FreeGo/Options/ShowVideo", cbShowVideo->toggled() );
	Options::current()->setValue( "/FreeGo/Options/RecordPath", fgDirLabel->text() );
}
